@@ -4,7 +4,7 @@
 MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st) {
     MVMObject *HOW = st->HOW;
     if (!HOW && st->HOW_sc)
-        st->HOW = HOW = MVM_sc_get_object(tc, st->HOW_sc, st->HOW_idx);
+        MVM_ASSIGN_REF(tc, &(st->header), st->HOW, HOW = MVM_sc_get_object(tc, st->HOW_sc, st->HOW_idx));
     return HOW ? HOW : tc->instance->VMNull;
 }
 
@@ -61,6 +61,9 @@ static void die_over_missing_method(MVMThreadContext *tc, MVMObject *obj, MVMStr
             c_name, MVM_6model_get_debug_name(tc, obj));
     }
 }
+static void find_method_unwind(MVMThreadContext *tc, void *sr_data) {
+    MVM_free(sr_data);
+}
 static void late_bound_find_method_return(MVMThreadContext *tc, void *sr_data) {
     FindMethodSRData *fm = (FindMethodSRData *)sr_data;
     if (MVM_is_null(tc, fm->res->o) || !IS_CONCRETE(fm->res->o)) {
@@ -157,7 +160,7 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
         fm->res  = res;
         fm->throw_if_not_found = throw_if_not_found;
         MVM_frame_special_return(tc, tc->cur_frame, late_bound_find_method_return,
-            NULL, fm, mark_find_method_sr_data);
+            find_method_unwind, fm, mark_find_method_sr_data);
     }
     tc->cur_frame->args[0].o = HOW;
     tc->cur_frame->args[1].o = obj;
@@ -178,15 +181,13 @@ MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMS
     if (!MVM_is_null(tc, meth)) {
         /* Got it; cache. Must be careful due to threads
          * reading, races, etc. */
-        MVMStaticFrame *sf = tc->cur_frame->static_info;
         uv_mutex_lock(&tc->instance->mutex_spesh_install);
         if (!tc->cur_frame->effective_spesh_slots[ss_idx + 1]) {
-            MVMStaticFrameSpesh *spesh = sf->body.spesh;
-            MVM_ASSIGN_REF(tc, &(spesh->common.header),
+            MVM_ASSIGN_REF(tc, &(tc->cur_frame->spesh_cand->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx + 1],
                            (MVMCollectable *)meth);
             MVM_barrier();
-            MVM_ASSIGN_REF(tc, &(spesh->common.header),
+            MVM_ASSIGN_REF(tc, &(tc->cur_frame->spesh_cand->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx],
                            (MVMCollectable *)STABLE(obj));
         }
@@ -329,6 +330,10 @@ static void mark_sr_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *w
     MVM_gc_worklist_add(tc, worklist, &atd->type);
 }
 
+static void free_sr_data(MVMThreadContext *tc, void *sr_data) {
+    MVM_free(sr_data);
+}
+
 void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MVMRegister *res) {
     MVMObject **cache;
     MVMSTable  *st;
@@ -387,7 +392,7 @@ void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MV
                 atd->obj = obj;
                 atd->type = type;
                 atd->res = res;
-                MVM_frame_special_return(tc, tc->cur_frame, accepts_type_sr, NULL,
+                MVM_frame_special_return(tc, tc->cur_frame, accepts_type_sr, free_sr_data,
                     atd, mark_sr_data);
             }
             STABLE(code)->invoke(tc, code, typecheck_callsite, tc->cur_frame->args);
@@ -475,7 +480,7 @@ MVMuint64 MVM_6model_next_type_cache_id(MVMThreadContext *tc) {
  * instances, marks the individual ojbect as never repossessable. */
 void MVM_6model_never_repossess(MVMThreadContext *tc, MVMObject *obj) {
     if (IS_CONCRETE(obj))
-        obj->header.flags |= MVM_CF_NEVER_REPOSSESS;
+        obj->header.flags1 |= MVM_CF_NEVER_REPOSSESS;
     else
         obj->st->mode_flags |= MVM_NEVER_REPOSSESS_TYPE;
 }
