@@ -3,6 +3,8 @@
 /* This representation's function pointer table. */
 static const MVMREPROps CPPStruct_this_repr;
 
+static void free_repr_data(MVMCPPStructREPRData *repr_data);
+
 /* Locates all of the attributes. Puts them onto a flattened, ordered
  * list of attributes (populating the passed flat_list). Also builds
  * the index mapping for doing named lookups. Note index is not related
@@ -16,17 +18,11 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
 
     MVMint32 mro_idx = MVM_repr_elems(tc, mro);
 
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&st);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&mro);
-
     flat_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&flat_list);
 
     class_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&class_list);
 
     attr_map_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map_list);
 
     /* Walk through the parents list. */
     while (mro_idx)
@@ -45,9 +41,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
             MVMObject *attr_map = NULL;
 
             if (MVM_iter_istrue(tc, attr_iter)) {
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_iter);
                 attr_map = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_hash_type);
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map);
             }
 
             while (MVM_iter_istrue(tc, attr_iter)) {
@@ -72,22 +66,17 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
                 MVM_repr_push_o(tc, flat_list, attr);
             }
 
-            if (attr_map) {
-                MVM_gc_root_temp_pop_n(tc, 2);
-            }
-
             /* Add to class list and map list. */
             MVM_repr_push_o(tc, class_list, current_class);
             MVM_repr_push_o(tc, attr_map_list, attr_map);
         }
         else {
+            MVM_gc_allocate_gen2_default_clear(tc);
             MVM_exception_throw_adhoc(tc,
                 "CPPStruct representation does not support multiple inheritance");
         }
     }
 
-    MVM_gc_root_temp_pop_n(tc, 5);
-
     /* We can now form the name map. */
     num_classes = MVM_repr_elems(tc, class_list);
     result = (MVMCPPStructNameMap *) MVM_malloc(sizeof(MVMCPPStructNameMap) * (1 + num_classes));
@@ -165,6 +154,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
             MVMint32   align = ALIGNOF(void *);
 
             if (num_dimensions > 1) {
+                MVM_gc_allocate_gen2_default_clear(tc);
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc,
                     "Only one dimensions supported in CPPStruct attribute");
             }
@@ -219,6 +210,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     if (inlined) {
                         MVMCArrayREPRData *carray_repr_data = (MVMCArrayREPRData *)STABLE(type)->REPR_data;
                         if (!carray_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CPPStruct: can't inline a CArray attribute before its type's definition");
                         }
@@ -250,6 +243,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         if (!cstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CPPStruct: can't inline a CStruct attribute before its type's definition");
                         }
@@ -266,6 +261,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     if (inlined) {
                         MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
                         if (!cppstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CPPStruct: can't inline a CPPStruct attribute before its type's definition");
                         }
@@ -282,6 +279,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         if (!cunion_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CPPStruct: can't inline a CUnion attribute before its type's definition");
                         }
@@ -297,6 +296,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                 }
                 else {
+                    MVM_gc_allocate_gen2_default_clear(tc);
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "CPPStruct representation only handles attributes of type:\n"
                         "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
@@ -304,12 +305,16 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                 }
             }
             else {
+                MVM_gc_allocate_gen2_default_clear(tc);
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc,
                     "CPPStruct representation requires the types of all attributes to be specified");
             }
 
             if (bits % 8) {
-                 MVM_exception_throw_adhoc(tc,
+                MVM_gc_allocate_gen2_default_clear(tc);
+                free_repr_data(repr_data);
+                MVM_exception_throw_adhoc(tc,
                     "CPPStruct only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
 
@@ -384,9 +389,11 @@ static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
 /* Composes the representation. */
 static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *repr_info) {
     /* Compute allocation strategy. */
-    MVMCPPStructREPRData *repr_data = MVM_calloc(1, sizeof(MVMCPPStructREPRData));
     MVMObject *attr_info = MVM_repr_at_key_o(tc, repr_info, tc->instance->str_consts.attribute);
+    MVMCPPStructREPRData *repr_data = MVM_calloc(1, sizeof(MVMCPPStructREPRData));
+    MVM_gc_allocate_gen2_default_set(tc);
     compute_allocation_strategy(tc, st, attr_info, repr_data);
+    MVM_gc_allocate_gen2_default_clear(tc);
     st->REPR_data = repr_data;
 }
 
@@ -464,41 +471,42 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                     /* No cached object. */
                     void *cobj = get_ptr_at_offset(body->cppstruct, repr_data->struct_offsets[slot]);
                     if (cobj) {
-                        MVMObject **child_objs = body->child_objs;
-                        if (type == MVM_CPPSTRUCT_ATTR_CARRAY) {
-                            obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CPPSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CUNION) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cunion(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CPTR) {
-                            obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_STRING) {
-                            MVMROOT(tc, typeobj, {
-                                MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
-                                    cobj, strlen(cobj));
-                                obj = MVM_repr_box_str(tc, typeobj, str);
-                            });
-                        }
+                        MVMROOT(tc, root, {
+                            if (type == MVM_CPPSTRUCT_ATTR_CARRAY) {
+                                obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CSTRUCT) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cstruct(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CPPSTRUCT) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CUNION) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cunion(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CPTR) {
+                                obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_STRING) {
+                                MVMROOT(tc, typeobj, {
+                                    MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
+                                        cobj, strlen(cobj));
+                                    obj = MVM_repr_box_str(tc, typeobj, str);
+                                });
+                            }
+                        });
                         MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                     }
                     else {
@@ -700,10 +708,7 @@ static void gc_mark_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist
     }
 }
 
-/* Free representation data. */
-static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
-    MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)st->REPR_data;
-
+static void free_repr_data(MVMCPPStructREPRData *repr_data) {
     /* May not have survived to composition. */
     if (repr_data == NULL)
         return;
@@ -717,7 +722,13 @@ static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
         MVM_free(repr_data->initialize_slots);
     }
 
-    MVM_free(st->REPR_data);
+    MVM_free(repr_data);
+}
+
+/* Free representation data. */
+static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
+    MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)st->REPR_data;
+    free_repr_data(repr_data);
 }
 
 /* This is called to do any cleanup of resources when an object gets
