@@ -1,12 +1,12 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps MVMContinuation_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. Also sets the invocation protocol handler in the STable. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st = MVM_gc_allocate_stable(tc, &MVMContinuation_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -25,15 +25,17 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* Adds held objects to the GC worklist. */
 static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMContinuationBody *body = (MVMContinuationBody *)data;
-    MVM_gc_worklist_add_frame(tc, worklist, body->top);
-    MVM_gc_worklist_add_frame(tc, worklist, body->root);
+    MVM_gc_worklist_add(tc, worklist, &body->top);
+    MVM_gc_worklist_add(tc, worklist, &body->root);
     if (body->active_handlers) {
         MVMActiveHandler *cur_ah = body->active_handlers;
         while (cur_ah != NULL) {
             MVM_gc_worklist_add(tc, worklist, &cur_ah->ex_obj);
+            MVM_gc_worklist_add(tc, worklist, &cur_ah->frame);
             cur_ah = cur_ah->next_handler;
         }
     }
+    MVM_gc_worklist_add(tc, worklist, &body->protected_tag);
     if (body->prof_cont) {
         MVMuint64 i;
         for (i = 0; i < body->prof_cont->num_sfs; i++)
@@ -44,15 +46,10 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
 /* Called by the VM in order to free memory associated with this object. */
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMContinuation *ctx = (MVMContinuation *)obj;
-    if (ctx->body.top)
-        ctx->body.top = MVM_frame_dec_ref(tc, ctx->body.top);
-    if (ctx->body.root)
-        ctx->body.root = MVM_frame_dec_ref(tc, ctx->body.root);
     if (ctx->body.active_handlers) {
         MVMActiveHandler *cur_ah = ctx->body.active_handlers;
         while (cur_ah != NULL) {
             MVMActiveHandler *next_ah = cur_ah->next_handler;
-            MVM_frame_dec_ref(tc, cur_ah->frame);
             MVM_free(cur_ah);
             cur_ah = next_ah;
         }
@@ -82,10 +79,10 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info) {
 
 /* Initializes the representation. */
 const MVMREPROps * MVMContinuation_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &MVMContinuation_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps MVMContinuation_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -111,7 +108,6 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "MVMContinuation", /* name */
     MVM_REPR_ID_MVMContinuation,
-    1, /* refs_frames */
     NULL, /* unmanaged_size */
     NULL, /* describe_refs */
 };
