@@ -18,6 +18,9 @@ static void exit_single_user(MVMThreadContext *tc, MVMArrayBody *arr) {
 #endif
 }
 
+#define MVM_MAX(a,b) ((a)>(b)?(a):(b))
+#define MVM_MIN(a,b) ((a)<(b)?(a):(b))
+
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
@@ -68,6 +71,11 @@ static void VMArray_gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVM
     MVMuint64         elems     = body->elems;
     MVMuint64         start     = body->start;
     MVMuint64         i         = 0;
+
+    /* Aren't holding anything, nothing to do. */
+    if (elems == 0)
+        return;
+
     switch (repr_data->slot_type) {
         case MVM_ARRAY_OBJ: {
             MVMObject **slots = body->slots.o;
@@ -626,12 +634,15 @@ static void unshift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *
     MVMArrayREPRData *repr_data = (MVMArrayREPRData *)st->REPR_data;
     MVMArrayBody     *body      = (MVMArrayBody *)data;
 
-    /* If we don't have room at the beginning of the slots,
-     * make some room (8 slots) for unshifting */
+    /* If we don't have room at the beginning of the slots, make some
+     * room for unshifting. We make room for a minimum of 8 elements, but
+     * for cases where we're just continuously unshifting factor in the
+     * body size too - however also apply an upper limit on that as in the
+     * push-based growth. */
     enter_single_user(tc, body);
     if (body->start < 1) {
-        MVMuint64 n = 8;
         MVMuint64 elems = body->elems;
+        MVMuint64 n = MVM_MIN(MVM_MAX(elems, 8), 8192);
 
         /* grow the array */
         set_size_internal(tc, body, elems + n, repr_data);
@@ -892,7 +903,9 @@ static void write_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
     }
 
     /* resize the array if necessary*/
-    if (elems < offset + count)
+    size_t elem_size = repr_data->elem_size;
+    /* No need to account for start, set_size_internal will take care of that */
+    if (elems * elem_size < offset * elem_size + count)
         set_size_internal(tc, body, offset + count, repr_data);
 
     memcpy(body->slots.u8 + (start + offset) * repr_data->elem_size, from, count);
@@ -903,8 +916,9 @@ MVMint64 read_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     MVMArrayBody     *body      = (MVMArrayBody *)data;
     MVMint64 start = body->start;
     MVMint64 result = 0;
+    size_t elem_size = repr_data->elem_size;
 
-    if (offset < 0 || start + body->elems < offset + count) {
+    if (offset < 0 || (start + body->elems) * elem_size < offset * elem_size + count) {
         MVM_exception_throw_adhoc(tc, "MVMArray: read_buf out of bounds offset %"PRIi64" start %"PRIi64" elems %"PRIu64" count %"PRIu64, offset, start, body->elems, count);
     }
 
