@@ -73,8 +73,8 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
 
     /* mostly stolen from validation.c */
     MVMStaticFrame *static_frame = frame;
-    MVMuint32 bytecode_size = maybe_candidate ? maybe_candidate->bytecode_size : static_frame->body.bytecode_size;
-    MVMuint8 *bytecode_start = maybe_candidate ? maybe_candidate->bytecode : static_frame->body.bytecode;
+    MVMuint32 bytecode_size = maybe_candidate ? maybe_candidate->body.bytecode_size : static_frame->body.bytecode_size;
+    MVMuint8 *bytecode_start = maybe_candidate ? maybe_candidate->body.bytecode : static_frame->body.bytecode;
     MVMuint8 *bytecode_end = bytecode_start + bytecode_size;
     /* current position in the bytestream */
     MVMuint8 *cur_op = bytecode_start;
@@ -257,11 +257,11 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
             }
             else if (op_rw == MVM_operand_read_reg || op_rw == MVM_operand_write_reg) {
                 /* register operand */
-                MVMuint8 frame_has_inlines = maybe_candidate && maybe_candidate->num_inlines ? 1 : 0;
-                MVMuint16 *local_types = frame_has_inlines ? maybe_candidate->local_types : frame->body.local_types;
+                MVMuint8 frame_has_inlines = maybe_candidate && maybe_candidate->body.num_inlines ? 1 : 0;
+                MVMuint16 *local_types = frame_has_inlines ? maybe_candidate->body.local_types : frame->body.local_types;
                 operand_size = 2;
                 a("loc_%u_%s", GET_REG(cur_op, 0),
-                    get_typename(local_types[GET_REG(cur_op, 0)]));
+                    local_types ? get_typename(local_types[GET_REG(cur_op, 0)]) : "unknown");
             }
             else if (op_rw == MVM_operand_read_lex || op_rw == MVM_operand_write_lex) {
                 /* lexical operand */
@@ -280,9 +280,15 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
                 /* inefficient, I know. should use a hash. */
                 for (m = 0; m < cu->body.num_frames; m++) {
                     if (get_frame(tc, cu, m) == applicable_frame) {
-                        char *lexname = frame_lexicals ? frame_lexicals[m][idx] : "lex??";
-                        a("lex_Frame_%u_%s_%s", m, lexname,
-                            get_typename(applicable_frame->body.lexical_types[idx]));
+                        if (frame_lexicals) {
+                            char *lexname = frame_lexicals[m][idx];
+                            a("lex_Frame_%u_%s_%s", m, lexname,
+                                get_typename(applicable_frame->body.lexical_types[idx]));
+                        }
+                        else {
+                            a("lex_Frame_%u_lex%d_%s", m, idx,
+                                get_typename(applicable_frame->body.lexical_types[idx]));
+                        }
                     }
                 }
             }
@@ -366,7 +372,6 @@ char * MVM_bytecode_dump(MVMThreadContext *tc, MVMCompUnit *cu) {
     MVMuint32 i, j, k;
     char *o = MVM_calloc(s, sizeof(char));
     char ***frame_lexicals = MVM_malloc(sizeof(char **) * cu->body.num_frames);
-    MVMString *name = MVM_string_utf8_decode(tc, tc->instance->VMString, "", 0);
 
     a("\nMoarVM dump of binary compilation unit:\n\n");
 
@@ -420,21 +425,24 @@ char * MVM_bytecode_dump(MVMThreadContext *tc, MVMCompUnit *cu) {
 
     for (k = 0; k < cu->body.num_frames; k++) {
         MVMStaticFrame *frame = get_frame(tc, cu, k);
-        MVMLexicalRegistry *current;
-        char **lexicals;
 
         if (!frame->body.fully_deserialized) {
             MVM_bytecode_finish_frame(tc, cu, frame, 1);
         }
 
-        lexicals = (char **)MVM_malloc(sizeof(char *) * frame->body.num_lexicals);
-        frame_lexicals[k] = lexicals;
+        MVMuint32 num_lexicals = frame->body.num_lexicals;
+        if (num_lexicals) {
+            MVMString **lexical_names_list = frame->body.lexical_names_list;
 
-        HASH_ITER(tc, hash_handle, frame->body.lexical_names, current, {
-            name->body.storage.blob_32 = (MVMint32 *)current->hash_handle.key;
-            name->body.num_graphs      = (MVMuint32)current->hash_handle.keylen / sizeof(MVMGrapheme32);
-            lexicals[current->value]   = MVM_string_utf8_encode_C_string(tc, name);
-        });
+            char **lexicals = (char **)MVM_malloc(sizeof(char *) * num_lexicals);
+            for (j = 0; j < num_lexicals; j++) {
+                lexicals[j]   = MVM_string_utf8_encode_C_string(tc, lexical_names_list[j]);
+            }
+            frame_lexicals[k] = lexicals;
+        }
+        else {
+            frame_lexicals[k] = NULL;
+        }
     }
     for (k = 0; k < cu->body.num_frames; k++) {
         MVMStaticFrame *frame = get_frame(tc, cu, k);
@@ -528,7 +536,7 @@ void MVM_dump_bytecode(MVMThreadContext *tc) {
         /*MVMuint8 found = 0;*/
         /*for (spesh_cand_idx = 0; spesh_cand_idx < sf->body.num_spesh_candidates; spesh_cand_idx++) {*/
             /*MVMSpeshCandidate *cand = sf->body.spesh_candidates[spesh_cand_idx];*/
-            /*if (cand->bytecode == effective_bytecode) {*/
+            /*if (cand->body.bytecode == effective_bytecode) {*/
                 /*MVM_dump_bytecode_of(tc, tc->cur_frame, cand);*/
                 /*found = 1;*/
             /*}*/
@@ -556,7 +564,7 @@ void MVM_dump_bytecode_stackframe(MVMThreadContext *tc, MVMint32 depth) {
         MVMStaticFrameSpesh *spesh = sf->body.spesh;
         for (spesh_cand_idx = 0; spesh_cand_idx < spesh->body.num_spesh_candidates; spesh_cand_idx++) {
             MVMSpeshCandidate *cand = spesh->body.spesh_candidates[spesh_cand_idx];
-            if (cand->bytecode == effective_bytecode) {
+            if (cand->body.bytecode == effective_bytecode) {
                 MVM_dump_bytecode_of(tc, frame, cand);
             }
         }
