@@ -10,7 +10,7 @@ static int crash_on_error = 0;
 /* Function for getting effective (specialized or not) frame handlers. */
 MVM_STATIC_INLINE MVMFrameHandler * MVM_frame_effective_handlers(MVMFrame *f) {
     MVMSpeshCandidate *spesh_cand = f->spesh_cand;
-    return spesh_cand ? spesh_cand->handlers : f->static_info->body.handlers;
+    return spesh_cand ? spesh_cand->body.handlers : f->static_info->body.handlers;
 }
 
 /* Maps ID of exception category to its name. */
@@ -100,10 +100,10 @@ static MVMint32 search_frame_handlers_dyn(MVMThreadContext *tc, MVMFrame *f,
                                           MVMuint32 cat, MVMObject *payload,
                                           LocatedHandler *lh) {
     MVMuint32  i;
-    if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
-        MVMJitCode *jitcode = f->spesh_cand->jitcode;
+    if (f->spesh_cand && f->spesh_cand->body.jitcode && f->jit_entry_label) {
+        MVMJitCode *jitcode = f->spesh_cand->body.jitcode;
         void *current_position = MVM_jit_code_get_current_position(tc, jitcode, f);
-        MVMJitHandler    *jhs = f->spesh_cand->jitcode->handlers;
+        MVMJitHandler    *jhs = f->spesh_cand->body.jitcode->handlers;
         MVMFrameHandler  *fhs = MVM_frame_effective_handlers(f);
         for (i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, 0);
              i < jitcode->num_handlers;
@@ -117,7 +117,7 @@ static MVMint32 search_frame_handlers_dyn(MVMThreadContext *tc, MVMFrame *f,
         }
     } else {
         MVMuint32 num_handlers = f->spesh_cand
-            ? f->spesh_cand->num_handlers
+            ? f->spesh_cand->body.num_handlers
             : f->static_info->body.num_handlers;
         MVMuint32 pc;
         if (f == tc->cur_frame)
@@ -175,8 +175,8 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
     MVMuint32 i;
     MVMuint32 skipping = *skip_first_inlinee;
     MVMFrameHandler *fhs = MVM_frame_effective_handlers(f);
-    if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
-        MVMJitCode *jitcode = f->spesh_cand->jitcode;
+    if (f->spesh_cand && f->spesh_cand->body.jitcode && f->jit_entry_label) {
+        MVMJitCode *jitcode = f->spesh_cand->body.jitcode;
         void *current_position = MVM_jit_code_get_current_position(tc, jitcode, f);
         MVMJitHandler    *jhs = jitcode->handlers;
         for (i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, 0);
@@ -191,7 +191,7 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
                     *skip_first_inlinee = 0;
                 }
                 else {
-                    MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
+                    MVMuint16 cr_reg = f->spesh_cand->body.inlines[fh->inlinee].code_ref_reg;
                     MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
                     if (inline_outer == f) {
                         skip_all_inlinees = 1;
@@ -215,7 +215,7 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
     }
     else {
         MVMuint32 num_handlers = f->spesh_cand
-            ? f->spesh_cand->num_handlers
+            ? f->spesh_cand->body.num_handlers
             : f->static_info->body.num_handlers;
         MVMuint32 pc;
         if (f == tc->cur_frame)
@@ -233,7 +233,7 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
                         *skip_first_inlinee = 0;
                     }
                     else {
-                        MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
+                        MVMuint16 cr_reg = f->spesh_cand->body.inlines[fh->inlinee].code_ref_reg;
                         MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
                         if (inline_outer == f) {
                             skip_all_inlinees = 1;
@@ -310,7 +310,7 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
                     lh.frame = f;
                     return lh;
                 }
-                f = f->caller;
+                f = MVM_LIKELY(f != tc->thread_entry_frame) ? f->caller : NULL;
             }
             return lh;
         case MVM_EX_THROW_LEXOTIC:
@@ -318,7 +318,7 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
                 lh = search_for_handler_from(tc, f, MVM_EX_THROW_LEX, cat, payload);
                 if (lh.frame != NULL)
                     return lh;
-                f = f->caller;
+                f = MVM_LIKELY(f != tc->thread_entry_frame) ? f->caller : NULL;
             }
             return lh;
         default:
@@ -348,8 +348,8 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
 
     case MVM_EX_ACTION_GOTO:
         if (lh.jit_handler) {
-            void **labels = lh.frame->spesh_cand->jitcode->labels;
-            MVMuint8  *pc = lh.frame->spesh_cand->jitcode->bytecode;
+            void **labels = lh.frame->spesh_cand->body.jitcode->labels;
+            MVMuint8  *pc = lh.frame->spesh_cand->body.jitcode->bytecode;
             MVM_frame_unwind_to(tc, lh.frame, pc, 0, NULL, labels[lh.jit_handler->goto_label]);
         } else {
             MVM_frame_unwind_to(tc, lh.frame, NULL, lh.handler->goto_offset, NULL, NULL);
@@ -396,6 +396,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
          * handler. */
         cur_frame->return_value = (MVMRegister *)&tc->last_handler_result;
         cur_frame->return_type = MVM_RETURN_OBJ;
+        cur_frame->return_address = *(tc->interp_cur_op);
         MVM_frame_special_return(tc, cur_frame, unwind_after_handler, cleanup_active_handler,
             ah, NULL);
 
@@ -427,9 +428,9 @@ static void unwind_after_handler(MVMThreadContext *tc, void *sr_data) {
     frame       = ah->frame;
     exception   = (MVMException *)ah->ex_obj;
     if (ah->jit_handler) {
-        void **labels = frame->spesh_cand->jitcode->labels;
+        void **labels = frame->spesh_cand->body.jitcode->labels;
         jit_return_label = labels[ah->jit_handler->goto_label];
-        abs_address = frame->spesh_cand->jitcode->bytecode;
+        abs_address = frame->spesh_cand->body.jitcode->bytecode;
         goto_offset = 0;
     }
     else {
@@ -572,7 +573,8 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
 
         /* row will contain "sub" and "annotations" */
         row = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
-        MVM_repr_bind_key_o(tc, row, k_sub, cur_frame->code_ref);
+        MVM_repr_bind_key_o(tc, row, k_sub,
+            cur_frame->code_ref ? cur_frame->code_ref : tc->instance->VMNull);
         MVM_repr_bind_key_o(tc, row, k_anno, annotations);
 
         MVM_repr_push_o(tc, arr, row);
@@ -602,20 +604,21 @@ MVMObject * MVM_exception_backtrace_strings(MVMThreadContext *tc, MVMObject *ex_
 
     MVMROOT(tc, ex, {
         arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    });
-    cur_frame = ex->body.origin;
 
-    MVMROOT2(tc, arr, cur_frame, {
-        MVMuint32 count = 0;
-        while (cur_frame != NULL) {
-            char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
-                ex->body.throw_address);
-            MVMString *line_str = MVM_string_utf8_decode(tc, tc->instance->VMString, line, strlen(line));
-            MVMObject *line_obj = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, line_str);
-            MVM_repr_push_o(tc, arr, line_obj);
-            cur_frame = cur_frame->caller;
-            MVM_free(line);
-        }
+        cur_frame = ex->body.origin;
+
+        MVMROOT2(tc, arr, cur_frame, {
+            MVMuint32 count = 0;
+            while (cur_frame != NULL) {
+                char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
+                    ex->body.throw_address);
+                MVMString *line_str = MVM_string_utf8_decode(tc, tc->instance->VMString, line, strlen(line));
+                MVMObject *line_obj = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, line_str);
+                MVM_repr_push_o(tc, arr, line_obj);
+                cur_frame = cur_frame->caller;
+                MVM_free(line);
+            }
+        });
     });
 
     return arr;
@@ -644,7 +647,13 @@ static void panic_unhandled_cat(MVMThreadContext *tc, MVMuint32 cat) {
             cat_name(tc, cat));
     }
     else {
-        fprintf(stderr, "No exception handler located for %s\n", cat_name(tc, cat));
+        if (tc->nested_interpreter)
+            fputs("An unhandled exception occurred in a native callback.\n"
+                  "This situation is not recoverable, and the program will terminate.\n"
+                  "The stack trace below helps indicate which library needs fixing.\n"
+                  "The exception was thrown at:\n", stderr);
+        else
+            fprintf(stderr, "No exception handler located for %s\n", cat_name(tc, cat));
         MVM_dump_backtrace(tc);
         if (crash_on_error)
             abort();
@@ -658,7 +667,9 @@ static void panic_unhandled_ex(MVMThreadContext *tc, MVMException *ex) {
     char *backtrace;
 
     /* If a debug session is running, notify the client. */
-    MVM_debugserver_notify_unhandled_exception(tc, ex);
+    MVMROOT(tc, ex, {
+        MVM_debugserver_notify_unhandled_exception(tc, ex);
+    });
 
     /* If it's a control exception, try promoting it to a catch one; use
      * the category name. */
@@ -751,7 +762,7 @@ void MVM_exception_throwobj(MVMThreadContext *tc, MVMuint8 mode, MVMObject *ex_o
         ex->body.resume_addr = *tc->interp_cur_op;
         /* Ensure that we store label where the JIT should return, if any */
         if (tc->jit_return_address != NULL) {
-            ex->body.jit_resume_label = MVM_jit_code_get_current_position(tc, tc->cur_frame->spesh_cand->jitcode, tc->cur_frame);
+            ex->body.jit_resume_label = MVM_jit_code_get_current_position(tc, tc->cur_frame->spesh_cand->body.jitcode, tc->cur_frame);
         }
     }
     lh = search_for_handler_from(tc, tc->cur_frame, mode, ex->body.category, ex->body.payload);
@@ -832,11 +843,11 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
  */
 MVM_NO_RETURN void MVM_panic(MVMint32 exitCode, const char *messageFormat, ...) {
     va_list args;
-    fprintf(stderr, "MoarVM panic: ");
+    fputs("MoarVM panic: ", stderr);
     va_start(args, messageFormat);
     vfprintf(stderr, messageFormat, args);
     va_end(args);
-    fwrite("\n", 1, 1, stderr);
+    fputc('\n', stderr);
     if (crash_on_error)
         abort();
     else
@@ -851,13 +862,23 @@ MVM_NO_RETURN void MVM_panic_allocation_failed(size_t len) {
  * process to indicate that we've made an error */
 MVM_NO_RETURN void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) {
     va_list args;
-    fprintf(stderr, "MoarVM oops: ");
+    fprintf(stderr, "MoarVM oops%s: ",
+            !tc ? " with NULL tc" :
+            (MVMObject *) tc->thread_obj == tc->instance->spesh_thread
+            ? " in spesh thread" :
+            (MVMObject *) tc->thread_obj == tc->instance->event_loop_thread
+            ? " in event loop thread" : "");
     va_start(args, messageFormat);
     vfprintf(stderr, messageFormat, args);
     va_end(args);
-    fprintf(stderr, "\n");
+    fputc('\n', stderr);
+
+    /* Our caller is seriously buggy if tc is NULL */
+    if (!tc)
+        abort();
+
     MVM_dump_backtrace(tc);
-    fprintf(stderr, "\n");
+    fputc('\n', stderr);
     exit(1);
 }
 
@@ -888,6 +909,23 @@ MVM_NO_RETURN void MVM_exception_throw_adhoc_free(MVMThreadContext *tc, char **w
 MVM_NO_RETURN void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const char *messageFormat, va_list args) {
     LocatedHandler lh;
     MVMException *ex;
+    const char *special = !tc ? " with NULL tc"
+        : (MVMObject *) tc->thread_obj == tc->instance->spesh_thread ? " in spesh thread"
+        : (MVMObject *) tc->thread_obj == tc->instance->event_loop_thread ? " in event loop thread" : NULL;
+
+    if (special) {
+        fprintf(stderr, "MoarVM exception%s treated as oops: ", special);
+        vfprintf(stderr, messageFormat, args);
+        va_end(args);
+        fputc('\n', stderr);
+        if (!tc)
+            abort();
+
+        MVM_dump_backtrace(tc);
+        fputc('\n', stderr);
+        exit(1);
+    }
+
     /* The current frame will be assigned as the thrower of the exception, so
      * force it onto the heap before we begin. */
     if (tc->cur_frame)
@@ -931,7 +969,7 @@ MVM_NO_RETURN void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char
         if (crash_on_error) {
             /* Yes, abort. */
             vfprintf(stderr, messageFormat, args);
-            fwrite("\n", 1, 1, stderr);
+            fputc('\n', stderr);
             MVM_dump_backtrace(tc);
             abort();
         }
