@@ -13,7 +13,7 @@
 #define FSA_SIZE_DEBUG 0
 #if FSA_SIZE_DEBUG
 typedef struct {
-    MVMuint64 alloc_size;
+    size_t alloc_size;
     void *memory;
 } MVMFixedSizeAllocDebug;
 #endif
@@ -26,9 +26,12 @@ MVMFixedSizeAlloc * MVM_fixed_size_create(MVMThreadContext *tc) {
 #endif
     MVMFixedSizeAlloc *al = MVM_malloc(sizeof(MVMFixedSizeAlloc));
     al->size_classes = MVM_calloc(MVM_FSA_BINS, sizeof(MVMFixedSizeAllocSizeClass));
-    if ((init_stat = uv_mutex_init(&(al->complex_alloc_mutex))) < 0)
+    if ((init_stat = uv_mutex_init(&(al->complex_alloc_mutex))) < 0) {
+        MVM_free(al->size_classes);
+        MVM_free(al);
         MVM_exception_throw_adhoc(tc, "Failed to initialize mutex: %s",
             uv_strerror(init_stat));
+    }
     al->freelist_spin = 0;
     al->free_at_next_safepoint_overflows = NULL;
 
@@ -174,6 +177,12 @@ static void * alloc_from_global(MVMThreadContext *tc, MVMFixedSizeAlloc *al, MVM
     return alloc_slow_path(tc, al, bin);
 }
 void * MVM_fixed_size_alloc(MVMThreadContext *tc, MVMFixedSizeAlloc *al, size_t bytes) {
+    /* This doesn't work with the current code, and it implies the caller is
+     * mighty confused. Hence it's better to flag this up than revise the rest
+     * of this code to hide the real problem. */
+    if (!bytes)
+        MVM_oops(tc, "MVM_fixed_size_alloc request for 0 bytes");
+
 #if FSA_SIZE_DEBUG
     MVMFixedSizeAllocDebug *dbg = MVM_malloc(bytes + sizeof(MVMuint64));
     dbg->alloc_size = bytes;
@@ -231,10 +240,9 @@ void * MVM_fixed_size_realloc(MVMThreadContext *tc, MVMFixedSizeAlloc *al, void
 void * MVM_fixed_size_realloc_at_safepoint(MVMThreadContext *tc, MVMFixedSizeAlloc *al, void * p, size_t old_bytes, size_t new_bytes) {
 #if FSA_SIZE_DEBUG
     MVMFixedSizeAllocDebug *new_p = MVM_fixed_size_alloc(tc, al, new_bytes);
-    MVMFixedSizeAllocDebug *dbg = (MVMFixedSizeAllocDebug *)((char *)new_p - 8);
     memcpy(new_p, (char *)p, new_bytes > old_bytes ? old_bytes : new_bytes);
     MVM_fixed_size_free_at_safepoint(tc, al, old_bytes, p);
-    return &(dbg->memory);
+    return new_p;
 #else
     MVMuint32 old_bin = bin_for(old_bytes);
     MVMuint32 new_bin = bin_for(new_bytes);
@@ -286,15 +294,15 @@ void MVM_fixed_size_free(MVMThreadContext *tc, MVMFixedSizeAlloc *al, size_t byt
 #ifdef MVM_VALGRIND_SUPPORT
         if (RUNNING_ON_VALGRIND) {
             char command[128];
-            snprintf(&command, 128, "check_memory defined %p %d", dbg, bytes + 8);
+            snprintf(command, 128, "check_memory defined %p %zu", dbg, bytes + 8);
             VALGRIND_MONITOR_COMMAND(command);
-            VALGRIND_PRINTF_BACKTRACE("Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+            VALGRIND_PRINTF_BACKTRACE("Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
         }
         else {
-            MVM_panic(1, "Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+            MVM_panic(1, "Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
         }
 #else
-        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
 #endif
     }
     MVM_free(dbg);
@@ -333,16 +341,16 @@ void MVM_fixed_size_free_at_safepoint(MVMThreadContext *tc, MVMFixedSizeAlloc *a
     if (dbg->alloc_size != bytes) {
 #ifdef MVM_VALGRIND_SUPPORT
         if (RUNNING_ON_VALGRIND) {
-            char command[128]; snprintf(&command, 128, "check_memory defined %p %lu", dbg, bytes + 8); VALGRIND_MONITOR_COMMAND(command);
-            VALGRIND_PRINTF_BACKTRACE("Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+            char command[128]; snprintf(command, 128, "check_memory defined %p %zu", dbg, bytes + 8); VALGRIND_MONITOR_COMMAND(command);
+            VALGRIND_PRINTF_BACKTRACE("Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
         }
         else {
-            MVM_panic(1, "Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+            MVM_panic(1, "Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
         }
 #else
-        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
 #endif
-        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %lu, got %lu", dbg->alloc_size, bytes);
+        MVM_panic(1, "Fixed size allocator: wrong size in free: expected %zu, got %zu", dbg->alloc_size, bytes);
     }
     add_to_overflows_safepoint_free_list(tc, al, dbg);
 #else
