@@ -181,6 +181,7 @@ typedef struct {
     char             **args;
     uv_stream_t       *stdin_handle;
     MVMuint32          had_stdin_handle;
+    int                stdin_to_close;
     MVMuint32          seq_stdout;
     MVMuint32          seq_stderr;
     MVMuint32          seq_merge;
@@ -235,7 +236,7 @@ static void on_write(uv_write_t *req, int status) {
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
     MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
-    MVM_free(wi->req);
+    MVM_free_null(wi->req);
 }
 
 /* Does setup work for an asynchronous write. */
@@ -268,8 +269,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         /* Error; need to notify. */
         MVMROOT(tc, async_task, {
             MVMObject    *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-            MVMAsyncTask *t   = (MVMAsyncTask *)async_task;
-            MVM_repr_push_o(tc, arr, t->body.schedulee);
+            MVM_repr_push_o(tc, arr, ((MVMAsyncTask *)async_task)->body.schedulee);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
             MVMROOT(tc, arr, {
                 MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
@@ -284,12 +284,11 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
                     tc->instance->boot_types.BOOTStr, msg_str);
                 MVM_repr_push_o(tc, arr, msg_box);
             });
-            MVM_repr_push_o(tc, t->body.queue, arr);
+            MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
         });
 
         /* Cleanup handle. */
-        MVM_free(wi->req);
-        wi->req = NULL;
+        MVM_free_null(wi->req);
     }
 }
 
@@ -360,6 +359,11 @@ static void proc_async_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *
         MVM_gc_worklist_add(tc, worklist, &(apd->async_task));
 }
 
+static void proc_async_gc_free(MVMThreadContext *tc, MVMObject *root, void *data) {
+    if (data)
+        MVM_free(data);
+}
+
 /* Does an asynchronous close (since it must run on the event loop). */
 static void close_cb(uv_handle_t *handle) {
     MVM_free(handle);
@@ -458,7 +462,7 @@ static const MVMIOOps proc_op_table = {
     NULL,
     NULL,
     proc_async_gc_mark,
-    NULL
+    proc_async_gc_free
 };
 
 static void spawn_async_close(uv_handle_t *handle) {
@@ -501,6 +505,12 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
     close_stdin(tc, os_handle);
     uv_mutex_unlock(mutex);
 
+    /* Close any stdin we were bound to. */
+    if (si->stdin_to_close) {
+        close(si->stdin_to_close);
+        si->stdin_to_close = 0;
+    }
+
     /* Close handle. */
     uv_close((uv_handle_t *)req, spawn_async_close);
     ((MVMIOAsyncProcessData *)((MVMOSHandle *)si->handle)->body.data)->handle = NULL;
@@ -681,6 +691,8 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[0].flags   = UV_INHERIT_FD;
         process_stdio[0].data.fd = (int)MVM_repr_get_int(tc,
             MVM_repr_at_key_o(tc, si->callbacks, tc->instance->str_consts.stdin_fd));
+        if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdin_fd_close))
+            si->stdin_to_close = process_stdio[0].data.fd;
     }
     else {
         process_stdio[0].flags   = UV_INHERIT_FD;
@@ -742,7 +754,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     process_options.file        = si->prog;
     process_options.args        = si->args;
     process_options.cwd         = si->cwd;
-    process_options.flags       = UV_PROCESS_WINDOWS_HIDE;
+    process_options.flags       = UV_PROCESS_WINDOWS_HIDE | UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS;
     process_options.env         = si->env;
     process_options.stdio_count = 3;
     process_options.exit_cb     = async_spawn_on_exit;
@@ -801,6 +813,11 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
                 MVM_repr_push_o(tc, arr, msg_box);
                 MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
             }
+
+            if (si->stdin_to_close) {
+                close(si->stdin_to_close);
+                si->stdin_to_close = 0;
+            }
         });
 
         MVM_io_eventloop_remove_active_work(tc, &(si->work_idx));
@@ -932,9 +949,11 @@ static void spawn_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *workl
 static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
     if (data) {
         SpawnInfo *si = (SpawnInfo *)data;
+        if (si->prog) {
+            MVM_free_null(si->prog);
+        }
         if (si->cwd) {
-            MVM_free(si->cwd);
-            si->cwd = NULL;
+            MVM_free_null(si->cwd);
         }
         if (si->env) {
             MVMuint32 i;
@@ -946,8 +965,13 @@ static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
             MVMuint32 i = 0;
             while (si->args[i])
                 MVM_free(si->args[i++]);
-            MVM_free(si->args);
-            si->args = NULL;
+            MVM_free_null(si->args);
+        }
+        if (si->pipe_stdout) {
+            MVM_free_null(si->pipe_stdout);
+        }
+        if (si->pipe_stderr) {
+            MVM_free_null(si->pipe_stderr);
         }
         MVM_free(si);
     }
@@ -963,12 +987,13 @@ static const MVMAsyncTaskOps spawn_op_table = {
 };
 
 /* Spawn a process asynchronously. */
-MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObject *argv,
-                                 MVMString *cwd, MVMObject *env, MVMObject *callbacks) {
+MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMString *prog,
+                                 MVMObject *argv, MVMString *cwd, MVMObject *env,
+                                 MVMObject *callbacks) {
     MVMAsyncTask  *task;
     MVMOSHandle   *handle;
     SpawnInfo     *si;
-    char          *prog, *_cwd, **_env, **args;
+    char          *_prog, *_cwd, **_env, **args;
     MVMuint64      size, arg_size, i;
     MVMIter       *iter;
     MVMRegister    reg;
@@ -978,17 +1003,20 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
         MVM_exception_throw_adhoc(tc,
             "spawnprocasync target queue must have ConcBlockingQueue REPR");
 
-    /* Encode arguments, taking first as program name. */
+    /* Encode program name and arguments. */
     arg_size = MVM_repr_elems(tc, argv);
     if (arg_size < 1)
         MVM_exception_throw_adhoc(tc, "spawnprocasync must have first arg for program");
+
+    _prog = MVM_string_utf8_c8_encode_C_string(tc, prog);
+
+    /* We need a trailing NULL byte, thus the +1. */
     args = MVM_malloc((arg_size + 1) * sizeof(char *));
     for (i = 0; i < arg_size; i++) {
         REPR(argv)->pos_funcs.at_pos(tc, STABLE(argv), argv, OBJECT_BODY(argv), i, &reg, MVM_reg_obj);
         args[i] = MVM_string_utf8_c8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
     }
     args[arg_size] = NULL;
-    prog = args[0];
 
     /* Encode CWD. */
     _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
@@ -1015,7 +1043,7 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
         MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
         task->body.ops  = &spawn_op_table;
         si              = MVM_calloc(1, sizeof(SpawnInfo));
-        si->prog        = prog;
+        si->prog        = _prog;
         si->cwd         = _cwd;
         si->env         = _env;
         si->args        = args;
@@ -1081,12 +1109,7 @@ MVMnum64 MVM_proc_randscale_n(MVMThreadContext *tc, MVMnum64 scale) {
 
 /* seed random number generator */
 void MVM_proc_seed(MVMThreadContext *tc, MVMint64 seed) {
-    /* Seed our one, plus the normal C srand for libtommath. */
     tinymt64_init(tc->rand_state, (MVMuint64)seed);
-/* do not call srand if we are not using rand */
-#ifndef MP_USE_ALT_RAND
-    srand((MVMuint32)seed);
-#endif
 }
 
 /* gets the system time since the epoch truncated to integral seconds */
