@@ -51,6 +51,7 @@ void MVM_profile_heap_start(MVMThreadContext *tc, MVMObject *config) {
         MVM_repr_at_key_o(tc, config, tc->instance->str_consts.path));
 
     if (MVM_is_null(tc, (MVMObject*)path_str)) {
+        MVM_free(col);
         MVM_exception_throw_adhoc(tc, "Didn't specify a path for the heap snapshot profiler");
     }
 
@@ -59,6 +60,7 @@ void MVM_profile_heap_start(MVMThreadContext *tc, MVMObject *config) {
     col->fh = MVM_platform_fopen(path, "w");
 
     if (!col->fh) {
+        MVM_free(col);
         char *waste[2] = {path, NULL};
         MVM_exception_throw_adhoc_free(tc, waste, "Couldn't open heap snapshot target file %s: %s", path, strerror(errno));
     }
@@ -246,52 +248,36 @@ static void add_reference_vm_str(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
    add_reference(tc, ss, MVM_SNAPSHOT_REF_KIND_STRING, str_idx, to);
 }
 
-/* Adds an entry to the seen hash. */
-static void saw(MVMThreadContext *tc, MVMHeapSnapshotState *ss, void *addr, MVMuint64 idx) {
-    MVMHeapSnapshotSeen *seen = MVM_calloc(1, sizeof(MVMHeapSnapshotSeen));
-    seen->address = addr;
-    seen->idx = idx;
-    HASH_ADD_KEYPTR(hash_handle, ss->seen, &(seen->address), sizeof(void *), seen);
-}
-
-/* Checks for an entry in the seen hash. If we find an entry, write the index
- * into the index pointer passed. */
-static MVMuint32 seen(MVMThreadContext *tc, MVMHeapSnapshotState *ss, void *addr, MVMuint64 *idx) {
-    MVMHeapSnapshotSeen *entry;
-    HASH_FIND(hash_handle, ss->seen, &addr, sizeof(void *), entry);
-    if (entry) {
-        *idx = entry->idx;
-        return 1;
-    }
-    else {
-        return 0;
-    }
-}
-
 /* Gets the index of a collectable, either returning an existing index if we've
  * seen it before or adding it if not. */
 static MVMuint64 get_collectable_idx(MVMThreadContext *tc,
         MVMHeapSnapshotState *ss, MVMCollectable *collectable) {
-    MVMuint64 idx;
-    if (!seen(tc, ss, collectable, &idx)) {
-        if (collectable->flags & MVM_CF_STABLE) {
-            idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_STABLE, collectable);
-            ss->col->total_stables++;
-        }
-        else if (collectable->flags & MVM_CF_TYPE_OBJECT) {
-            idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_TYPE_OBJECT, collectable);
-            ss->col->total_typeobjects++;
-        }
-        else if (collectable->flags & MVM_CF_FRAME) {
-            idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME, collectable);
-            ss->col->total_frames++;
-        }
-        else {
-            idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_OBJECT, collectable);
-            ss->col->total_objects++;
-        }
-        saw(tc, ss, collectable, idx);
+    struct MVMPtrHashEntry *entry = MVM_ptr_hash_lvalue_fetch(tc, &ss->seen, collectable);
+
+    if (entry->key) {
+        return entry->value;
     }
+    entry->key = collectable;
+
+    MVMuint64 idx;
+    if (collectable->flags1 & MVM_CF_STABLE) {
+        idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_STABLE, collectable);
+        ss->col->total_stables++;
+    }
+    else if (collectable->flags1 & MVM_CF_TYPE_OBJECT) {
+        idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_TYPE_OBJECT, collectable);
+        ss->col->total_typeobjects++;
+    }
+    else if (collectable->flags1 & MVM_CF_FRAME) {
+        idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME, collectable);
+        ss->col->total_frames++;
+    }
+    else {
+        idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_OBJECT, collectable);
+        ss->col->total_objects++;
+    }
+
+    entry->value = idx;
     return idx;
 }
 
@@ -299,12 +285,16 @@ static MVMuint64 get_collectable_idx(MVMThreadContext *tc,
  * it before or adding it if not. */
 static MVMuint64 get_frame_idx(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
         MVMFrame *frame) {
-    MVMuint64 idx;
-    if (!seen(tc, ss, frame, &idx)) {
-        idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME, frame);
-        ss->col->total_frames++;
-        saw(tc, ss, frame, idx);
+    struct MVMPtrHashEntry *entry = MVM_ptr_hash_lvalue_fetch(tc, &ss->seen, frame);
+
+    if (entry->key) {
+        return entry->value;
     }
+    entry->key = frame;
+
+    MVMuint64 idx = push_workitem(tc, ss, MVM_SNAPSHOT_COL_KIND_FRAME, frame);
+    ss->col->total_frames++;
+    entry->value = idx;
     return idx;
 }
 
@@ -371,9 +361,25 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
     MVMCompUnit *cu = sf->body.cu;
     MVMBytecodeAnnotation *ann = MVM_bytecode_resolve_annotation(tc, &(sf->body), 0);
     MVMuint64 line = ann ? ann->line_number : 1;
-    MVMuint64 file_idx = ann && ann->filename_string_heap_index < cu->body.num_strings
-        ? get_vm_string_index(tc, ss, MVM_cu_string(tc, cu, ann->filename_string_heap_index))
-        : get_vm_string_index(tc, ss, cu->body.filename);
+    MVMString *file_name = ann && ann->filename_string_heap_index < cu->body.num_strings
+        ? MVM_cu_string(tc, cu, ann->filename_string_heap_index)
+        : cu->body.filename;
+
+    /* We're running as part of gc_finalize. By now, live objects have been marked
+       but not yet collected. If the file name was just deserialized from the string
+       heap by MVM_cu_string, it will have been allocated in gen2 directly, but not
+       marked as live for this gc run. Do it here to prevent it from getting freed
+       after taking this heap snapshot, while it's actually still referenced from the
+       comp unit's strings list. But only do so if we're actually collecting gen2
+       in this run. Otherwise the flag may still be set during global destruction. */
+    if (
+        tc->instance->gc_full_collect
+        && file_name
+        && file_name->common.header.flags2 & MVM_CF_SECOND_GEN
+    )
+        file_name->common.header.flags2 |= MVM_CF_GEN2_LIVE;
+
+    MVMuint64 file_idx = get_vm_string_index(tc, ss, file_name);
 
     MVMuint64 i;
     MVMHeapSnapshotStaticFrame *s;
@@ -603,10 +609,10 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                     MVMuint16  i, count;
                     MVMuint16 *type_map;
                     MVMuint16  name_count = frame->static_info->body.num_lexicals;
-                    MVMLexicalRegistry **names = frame->static_info->body.lexical_names_list;
-                    if (frame->spesh_cand && frame->spesh_cand->lexical_types) {
-                        type_map = frame->spesh_cand->lexical_types;
-                        count    = frame->spesh_cand->num_lexicals;
+                    MVMString **names = frame->static_info->body.lexical_names_list;
+                    if (frame->spesh_cand && frame->spesh_cand->body.lexical_types) {
+                        type_map = frame->spesh_cand->body.lexical_types;
+                        count    = frame->spesh_cand->body.num_lexicals;
                     }
                     else {
                         type_map = frame->static_info->body.lexical_types;
@@ -616,7 +622,7 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                         if (type_map[i] == MVM_reg_str || type_map[i] == MVM_reg_obj) {
                             if (i < name_count)
                                 MVM_profile_heap_add_collectable_rel_vm_str(tc, ss,
-                                    (MVMCollectable *)frame->env[i].o, names[i]->key);
+                                    (MVMCollectable *)frame->env[i].o, names[i]);
                             else
                                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                                     (MVMCollectable *)frame->env[i].o, "Lexical (inlined)");
@@ -790,18 +796,27 @@ static void record_snapshot(MVMThreadContext *tc, MVMHeapSnapshotCollection *col
 
     /* Clean up temporary state. */
     MVM_free(ss.workitems);
-    MVM_HASH_DESTROY(tc, hash_handle, MVMHeapSnapshotSeen, ss.seen);
+    MVM_ptr_hash_demolish(tc, &ss.seen);
     MVM_gc_worklist_destroy(tc, ss.gcwl);
 }
 
 static void destroy_current_heap_snapshot(MVMThreadContext *tc) {
     MVMHeapSnapshotCollection *col = tc->instance->heap_snapshots;
 
+    MVM_free(col->snapshot->stats->type_counts);
+    MVM_free(col->snapshot->stats->type_size_sum);
+    MVM_free(col->snapshot->stats->sf_counts);
+    MVM_free(col->snapshot->stats->sf_size_sum);
+    MVM_free(col->snapshot->stats);
     MVM_free(col->snapshot->collectables);
     MVM_free(col->snapshot->references);
-    MVM_free(col->snapshot);
+    MVM_free_null(col->snapshot);
+}
 
-    col->snapshot = NULL;
+static void destroy_toc(MVMHeapDumpTableOfContents *toc) {
+    MVM_free(toc->toc_words);
+    MVM_free(toc->toc_positions);
+    MVM_free(toc);
 }
 
 /* Frees all memory associated with the heap snapshot. */
@@ -819,9 +834,9 @@ static void destroy_heap_snapshot_collection(MVMThreadContext *tc) {
     MVM_free(col->static_frames);
 
 #if MVM_HEAPSNAPSHOT_FORMAT == 3
-    MVM_free(col->toplevel_toc);
+    destroy_toc(col->toplevel_toc);
     if (col->second_level_toc)
-        MVM_free(col->second_level_toc);
+        destroy_toc(col->second_level_toc);
 #elif MVM_HEAPSNAPSHOT_FORMAT == 2
     MVM_free(col->index->snapshot_sizes);
     MVM_free(col->index);
@@ -892,7 +907,22 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
 
     {
         char namebuf[8];
+        /* Yes, this is a lot of boiler plate to silence a bogus warning :(
+         * Unfortunately, this is a real edge case. Using memcpy would lead to
+         * a buffer overflow if name is shorter than 8 bytes. The compiler
+         * warning aside, strncpy seems like exactly the right tool for the job
+         * as we want at most 8 bytes and don't care for any trailing \0, but
+         * are OK with zero padding at the end. */
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunknown-pragmas"
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wunknown-warning-option"
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wstringop-truncation"
         strncpy(namebuf, name, 8);
+#pragma GCC diagnostic pop
+#pragma clang diagnostic pop
+#pragma GCC diagnostic pop
         fwrite(namebuf, 8, 1, fh);
     }
 
@@ -1002,8 +1032,10 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
         i++;
     }
 
-    if (result_buffer_insert_pos == result_buffer)
+    if (result_buffer_insert_pos == result_buffer) {
+        MVM_free(result_buffer_insert_pos);
         return;
+    }
 
     size_position = ftell(fh);
     fwrite(&typename, sizeof(char), 8, fh);
@@ -1013,6 +1045,7 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
     cstream = ZSTD_createCStream();
 
     if (ZSTD_isError(return_value = ZSTD_initCStream(cstream, ZSTD_COMPRESSION_VALUE))) {
+        MVM_free(result_buffer);
         MVM_panic(1, "ZSTD compression error in heap snapshot: %s", ZSTD_getErrorName(return_value));
     }
 
@@ -1034,6 +1067,8 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
         return_value = ZSTD_compressStream(cstream, &outbuf, &inbuf);
 
         if (ZSTD_isError(return_value)) {
+            MVM_free(outbuf.dst);
+            MVM_free(result_buffer);
             MVM_panic(1, "Error compressing heap snapshot data: %s", ZSTD_getErrorName(return_value));
         }
 
@@ -1060,6 +1095,8 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
 
 
     if (ZSTD_isError(return_value)) {
+        MVM_free(outbuf.dst);
+        MVM_free(result_buffer);
         MVM_panic(1, "Error compressing heap snapshot data: %s", ZSTD_getErrorName(return_value));
     }
 
@@ -1345,7 +1382,7 @@ void snapshot_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollection
     MVMHeapDumpIndexSnapshotEntry *entry = NULL;
 
     MVMHeapDumpTableOfContents *outer_toc = col->toplevel_toc;
-    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(sizeof(MVMHeapDumpTableOfContents), 1);
+    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(1, sizeof(MVMHeapDumpTableOfContents));
 
     /*fprintf(stderr, "creating a new TOC at %p\n", inner_toc);*/
 
@@ -1758,6 +1795,7 @@ static void filemeta_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCol
     fwrite(&size, sizeof(MVMuint64), 1, fh);
 
     fputs(metadata, fh);
+    MVM_free(metadata);
     fputc(0, fh);
 
     end_position = ftell(fh);
@@ -1809,6 +1847,7 @@ static void snapmeta_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCol
     fwrite(&size, sizeof(MVMuint64), 1, fh);
 
     fputs(metadata, fh);
+    MVM_free(metadata);
     fputc(0, fh);
 
     end_position = ftell(fh);
@@ -1829,7 +1868,7 @@ void finish_collection_to_filehandle(MVMThreadContext *tc, MVMHeapSnapshotCollec
 
 #if MVM_HEAPSNAPSHOT_FORMAT == 3
     {
-    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(sizeof(MVMHeapDumpTableOfContents), 1);
+    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(1, sizeof(MVMHeapDumpTableOfContents));
 
     /*fprintf(stderr, "finishing collection; creating a new TOC at %p\n", inner_toc);*/
 
