@@ -10,9 +10,24 @@
 #define GET_UI8(pc, idx)    *((MVMuint8 *)((pc) + (idx)))
 #define GET_I16(pc, idx)    *((MVMint16 *)((pc) + (idx)))
 #define GET_UI16(pc, idx)   *((MVMuint16 *)((pc) + (idx)))
-#define GET_I32(pc, idx)    *((MVMint32 *)((pc) + (idx)))
-#define GET_UI32(pc, idx)   *((MVMuint32 *)((pc) + (idx)))
-#define GET_N32(pc, idx)    *((MVMnum32 *)((pc) + (idx)))
+
+MVM_STATIC_INLINE MVMint32 GET_I32(const MVMuint8 *pc, MVMint32 idx) {
+    MVMint32 retval;
+    memcpy(&retval, pc + idx, sizeof(retval));
+    return retval;
+}
+
+MVM_STATIC_INLINE MVMuint32 GET_UI32(const MVMuint8 *pc, MVMint32 idx) {
+    MVMuint32 retval;
+    memcpy(&retval, pc + idx, sizeof(retval));
+    return retval;
+}
+
+MVM_STATIC_INLINE MVMuint32 GET_N32(const MVMuint8 *pc, MVMint32 idx) {
+    MVMnum32 retval;
+    memcpy(&retval, pc + idx, sizeof(retval));
+    return retval;
+}
 
 /* Allocate a piece of memory from the spesh graph's region
  * allocator. Deallocated when the spesh graph is. */
@@ -47,7 +62,7 @@ void MVM_spesh_graph_grow_deopt_table(MVMThreadContext *tc, MVMSpeshGraph *g) {
 }
 
 /* Records a de-optimization annotation and mapping pair. */
-void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g,
+MVMint32 MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g,
                                           MVMSpeshIns *ins_node, MVMuint32 deopt_target,
                                           MVMint32 type) {
     /* Add an annotations. */
@@ -61,6 +76,7 @@ void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g
     MVM_spesh_graph_grow_deopt_table(tc, g);
     g->deopt_addrs[2 * g->num_deopt_addrs] = deopt_target;
     g->num_deopt_addrs++;
+    return ann->data.deopt_idx;
 }
 
 MVM_FORMAT(printf, 4, 5)
@@ -83,10 +99,12 @@ MVM_PUBLIC void MVM_spesh_graph_add_comment(MVMThreadContext *tc, MVMSpeshGraph
 
     ann               = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
     ann->type         = MVM_SPESH_ANN_COMMENT;
-    ann->data.comment = comment;
     ann->next         = ins->annotations;
     ins->annotations  = ann;
 
+    ann->data.comment  = comment;
+    ann->order = g->next_annotation_idx++;
+
     va_start(ap, fmt);
     vsnprintf(comment, size, fmt, ap);
     va_end(ap);
@@ -173,13 +191,12 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     MVMuint8    *end      = g->bytecode + g->bytecode_size;
     MVMuint32    ins_idx  = 0;
     MVMuint8     next_bbs = 1; /* Next iteration (here, first) starts a BB. */
-    MVMuint32    lineno_ann_offs = 0;
     MVMuint32    num_osr_points = 0;
 
     MVMBytecodeAnnotation *ann_ptr = MVM_bytecode_resolve_annotation(tc, &sf->body, sf->body.bytecode - pc);
 
     for (i = 0; i < g->num_handlers; i++) {
-        if (g->handlers[i].start_offset != -1 && g->handlers[i].goto_offset != -1) {
+        if (g->handlers[i].start_offset != (MVMuint32)-1 && g->handlers[i].goto_offset != (MVMuint32)-1) {
             byte_to_ins_flags[g->handlers[i].start_offset] |= MVM_CFG_BB_START;
             byte_to_ins_flags[g->handlers[i].end_offset] |= MVM_CFG_BB_START;
             byte_to_ins_flags[g->handlers[i].goto_offset] |= MVM_CFG_BB_START;
@@ -326,10 +343,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                         (int)type, ins_node->info->name);
                 }
                 break;
+            }
             default:
                 break;
             }
-            }
         }
 
         /* We specially handle the jumplist case, which needs to mark all of
@@ -411,7 +428,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         /* Start or got may be -1 if the code the handler covered became
          * dead. If so, mark the handler as removed. Ditto if end is
          * before start (would never match). */
-        if (g->handlers[i].start_offset == -1 || g->handlers[i].goto_offset == -1 ||
+        if (g->handlers[i].start_offset == (MVMuint32)-1 || g->handlers[i].goto_offset == (MVMuint32)-1 ||
                 g->handlers[i].start_offset > g->handlers[i].end_offset) {
             if (!g->unreachable_handlers)
                 g->unreachable_handlers = MVM_spesh_alloc(tc, g, g->num_handlers);
@@ -555,7 +572,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             for (i = 0; i < g->num_handlers; i++) {
                 if (is_catch_handler(tc, g, i)) {
                     MVMuint32 offset = g->handlers[i].goto_offset;
-                    if (offset != -1)
+                    if (offset != (MVMuint32)-1)
                         cur_bb->succ[insert_pos++] = ins_to_bb[byte_to_ins_flags[offset] >> 3];
                 }
             }
@@ -829,7 +846,7 @@ MVMSpeshBB ** MVM_spesh_graph_reverse_postorder(MVMThreadContext *tc, MVMSpeshGr
 static void iter_check(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB **rpo, MVMint32 *doms, MVMint32 iters) {
     if (iters > 100000) {
 #ifdef NDEBUG
-        MVMint32 k;
+        MVMuint32 k;
         char *dump_msg = MVM_spesh_dump(tc, g);
         printf("%s", dump_msg);
         MVM_free(dump_msg);
@@ -863,7 +880,7 @@ static MVMint32 intersect(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB **r
 
 /* Computes dominator information about the basic blocks. */
 static MVMint32 * compute_dominators(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB **rpo) {
-    MVMint32 i, j, changed;
+    MVMuint32 i, j, changed;
 
     /* Create result list, with all initialized to undefined (use -1, as it's
      * not a valid basic block index). Start node dominates itself. */
@@ -897,7 +914,7 @@ static MVMint32 * compute_dominators(MVMThreadContext *tc, MVMSpeshGraph *g, MVM
                 MVM_oops(tc, "Spesh: could not find processed initial dominator");
             }
             for (j = 0; j < b->num_pred; j++) {
-                if (j != chosen_pred) {
+                if (j != (MVMuint32)chosen_pred) {
                     MVMint32 p_idx = b->pred[j]->rpo_idx;
                     if (doms[p_idx] != -1)
                         new_idom = intersect(tc, g, rpo, doms, p_idx, new_idom);
@@ -932,11 +949,11 @@ static void add_child(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *target
     target->num_children++;
 }
 static void add_children(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB **rpo, MVMint32 *doms) {
-    MVMint32 i;
+    MVMuint32 i;
     for (i = 0; i < g->num_bbs; i++) {
         MVMSpeshBB *bb   = rpo[i];
-        MVMint32    idom = doms[i];
-        if (idom != i)
+        MVMuint32   idom = doms[i];
+        if ((MVMuint32)idom != i)
             add_child(tc, g, rpo[idom], bb);
     }
 }
@@ -1313,29 +1330,29 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     /* Create top-level graph object. */
     MVMSpeshGraph *g     = MVM_calloc(1, sizeof(MVMSpeshGraph));
     g->sf                = sf;
-    g->bytecode          = cand->bytecode;
-    g->bytecode_size     = cand->bytecode_size;
-    g->handlers          = cand->handlers;
-    g->num_handlers      = cand->num_handlers;
-    g->num_locals        = cand->num_locals;
-    g->num_lexicals      = cand->num_lexicals;
-    g->inlines           = cand->inlines;
-    g->num_inlines       = cand->num_inlines;
-    g->deopt_addrs       = cand->deopts;
-    g->num_deopt_addrs   = cand->num_deopts;
-    g->alloc_deopt_addrs = cand->num_deopts;
-    g->deopt_named_used_bit_field = cand->deopt_named_used_bit_field;
-    g->deopt_pea         = cand->deopt_pea;
-    g->local_types       = cand->local_types;
-    g->lexical_types     = cand->lexical_types;
-    g->num_spesh_slots   = cand->num_spesh_slots;
-    g->alloc_spesh_slots = cand->num_spesh_slots;
+    g->bytecode          = cand->body.bytecode;
+    g->bytecode_size     = cand->body.bytecode_size;
+    g->handlers          = cand->body.handlers;
+    g->num_handlers      = cand->body.num_handlers;
+    g->num_locals        = cand->body.num_locals;
+    g->num_lexicals      = cand->body.num_lexicals;
+    g->inlines           = cand->body.inlines;
+    g->num_inlines       = cand->body.num_inlines;
+    g->deopt_addrs       = cand->body.deopts;
+    g->num_deopt_addrs   = cand->body.num_deopts;
+    g->alloc_deopt_addrs = cand->body.num_deopts;
+    g->deopt_named_used_bit_field = cand->body.deopt_named_used_bit_field;
+    g->deopt_pea         = cand->body.deopt_pea;
+    g->local_types       = cand->body.local_types;
+    g->lexical_types     = cand->body.lexical_types;
+    g->num_spesh_slots   = cand->body.num_spesh_slots;
+    g->alloc_spesh_slots = cand->body.num_spesh_slots;
     g->phi_infos         = MVM_spesh_alloc(tc, g, MVMPhiNodeCacheSize * sizeof(MVMOpInfo));
     g->cand              = cand;
 
     g->spesh_slots       = MVM_malloc(g->alloc_spesh_slots * sizeof(MVMCollectable *));
 
-    memcpy(g->spesh_slots, cand->spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
+    memcpy(g->spesh_slots, cand->body.spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
 
     /* Ensure the frame is validated, since we'll rely on this. */
     if (sf->body.instrumentation_level == 0) {
@@ -1344,7 +1361,7 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts, cand->deopt_usage_info,
+    build_cfg(tc, g, sf, cand->body.deopts, cand->body.num_deopts, cand->body.deopt_usage_info,
             deopt_usage_ins_out);
     if (!cfg_only) {
         MVM_spesh_eliminate_dead_bbs(tc, g, 0);
@@ -1416,6 +1433,10 @@ void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist
     /* Mark inlines. */
     for (i = 0; i < g->num_inlines; i++)
         MVM_gc_worklist_add(tc, worklist, &(g->inlines[i].sf));
+
+    /* Mark spesh candidate. */
+    if (g->cand)
+        MVM_gc_worklist_add(tc, worklist, &(g->cand));
 }
 
 /* Describes GCables for a heap snapshot. */
@@ -1461,21 +1482,23 @@ void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* If there is a candidate that we either generated or that this graph was
      * generated from, it has ownership of the malloc'd memory. If not, then we
      * need to clean up */
-    if (g->spesh_slots && (!g->cand || g->cand->spesh_slots != g->spesh_slots))
+    if (g->spesh_slots && (!g->cand || g->cand->body.spesh_slots != g->spesh_slots))
         MVM_free(g->spesh_slots);
-    if (g->deopt_addrs && (!g->cand || g->cand->deopts != g->deopt_addrs))
+    if (g->deopt_addrs && (!g->cand || g->cand->body.deopts != g->deopt_addrs))
         MVM_free(g->deopt_addrs);
-    if (g->inlines && (!g->cand || g->cand->inlines != g->inlines))
+    if (g->inlines && (!g->cand || g->cand->body.inlines != g->inlines))
         MVM_free(g->inlines);
-    if (g->local_types &&  (!g->cand || g->cand->local_types != g->local_types))
+    if (g->local_types &&  (!g->cand || g->cand->body.local_types != g->local_types))
         MVM_free(g->local_types);
-    if (g->lexical_types &&  (!g->cand || g->cand->lexical_types != g->lexical_types))
+    if (g->lexical_types &&  (!g->cand || g->cand->body.lexical_types != g->lexical_types))
         MVM_free(g->lexical_types);
+    if (!g->cand)
+        MVM_spesh_pea_destroy_deopt_info(tc, &(g->deopt_pea));
 
     /* Handlers can come directly from static frame, from spesh candidate, and
      * from malloc/realloc. We only free it in the last case */
     if (g->handlers && g->handlers != g->sf->body.handlers &&
-        (!g->cand || g->cand->handlers != g->handlers))
+        (!g->cand || g->cand->body.handlers != g->handlers))
         MVM_free(g->handlers);
 
     /* Free the graph itself. */
