@@ -208,10 +208,10 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
 /* Gets the effective size for inlining considerations of a specialization,
  * which is its code size minus the code size of its inlines. */
 static MVMint32 get_effective_size(MVMThreadContext *tc, MVMSpeshCandidate *cand) {
-    MVMint32 result = cand->bytecode_size;
+    MVMint32 result = cand->body.bytecode_size;
     MVMuint32 i;
-    for (i = 0; i < cand->num_inlines; i++)
-        result -= cand->inlines[i].bytecode_size;
+    for (i = 0; i < cand->body.num_inlines; i++)
+        result -= cand->body.inlines[i].bytecode_size;
     if (result < 0)
         result = 0;
     return (MVMuint32)result;
@@ -277,7 +277,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
          * sure it stays available for deopt. */
         MVMuint32 i;
         MVM_spesh_facts_discover(tc, ig, NULL, 1);
-        add_deopt_usages(tc, ig, cand->deopt_usage_info, deopt_usage_ins);
+        add_deopt_usages(tc, ig, cand->body.deopt_usage_info, deopt_usage_ins);
         for (i = 0; i < ig->num_inlines; i++) {
             /* We can't be very precise about this, because we don't know the
              * SSA version in effect. So bump usages of all version of the
@@ -454,6 +454,17 @@ static void resize_handlers_table(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     }
 }
 
+static void rewrite_curcode(MVMThreadContext *tc, MVMSpeshGraph *g,
+                                 MVMSpeshIns *ins, MVMuint16 num_locals, MVMSpeshOperand code_ref_reg) {
+    MVMSpeshOperand *new_operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+    new_operands[0] = ins->operands[0];
+    new_operands[0].reg.orig += num_locals;
+    new_operands[1] = code_ref_reg;
+    ins->info = MVM_op_get_op(MVM_OP_set);
+    ins->operands = new_operands;
+    MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
+}
+
 /* Rewrites a lexical lookup to an outer to be done via. a register holding
  * the outer coderef. */
 static void rewrite_outer_lookup(MVMThreadContext *tc, MVMSpeshGraph *g,
@@ -613,6 +624,9 @@ static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 for (i = 0; i < ins->info->num_operands; i++)
                     ins->operands[i].reg.orig += inliner->num_locals;
             }
+            else if (opcode == MVM_OP_curcode) {
+                rewrite_curcode(tc, inliner, ins, inliner->num_locals, code_ref_reg);
+            }
             else if (opcode == MVM_OP_sp_getlex_o && ins->operands[1].lex.outers > 0) {
                 rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
                     MVM_OP_sp_getlexvia_o, code_ref_reg);
