@@ -383,11 +383,9 @@ void incorporate_stats(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
     /* Clear up offset logs and call type info; they're either incorproated or
      * to be tossed. Also zero OSR hits, so we don't over-inflate them if this
      * frame entry survives. */
-    MVM_free(simf->offset_logs);
-    simf->offset_logs = NULL;
+    MVM_free_null(simf->offset_logs);
     simf->offset_logs_used = simf->offset_logs_limit = 0;
-    MVM_free(simf->call_type_info);
-    simf->call_type_info = NULL;
+    MVM_free_null(simf->call_type_info);
     simf->call_type_info_used = simf->call_type_info_limit = 0;
     simf->osr_hits = 0;
 }
@@ -490,6 +488,10 @@ static void save_or_free_sim_stack(MVMThreadContext *tc, MVMSpeshSimStack *sims,
     if (first_survivor >= 0) {
         /* Move survivors to the start. */
         if (first_survivor > 0) {
+            for (i = 0; i < (MVMuint32)first_survivor; i++) {
+                MVM_free(sims->frames[i].offset_logs);
+                MVM_free(sims->frames[i].call_type_info);
+            }
             sims->used -= first_survivor;
             memmove(sims->frames, sims->frames + first_survivor,
                 sims->used * sizeof(MVMSpeshSimStackFrame));
